﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.IO;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Data;
using System.Windows.Input;
using System.Xml;

using CodeManager.Command;
using CodeManager.Entity;
using CodeManager.Model;
using CodeManager.Service;

using ICSharpCode.AvalonEdit.Document;

using MahApps.Metro.Controls.Dialogs;

using Ookii.Dialogs.Wpf;

namespace CodeManager.ViewModel
{
    public class MainWindowViewModel : ViewModelBase, IDataErrorInfo
    {
        #region 所需xml信息

        private readonly string snippetXml1 =
            "<?xml version=\"1.0\" encoding=\"utf-8\"?><CodeSnippets xmlns=\"http://schemas.microsoft.com/VisualStudio/2005/CodeSnippet\"><CodeSnippet Format=\"1.0.0\"><Header>";

        private readonly string snippetXml2 = "<SnippetTypes><SnippetType>Expansion</SnippetType><SnippetType>SurroundsWith</SnippetType></SnippetTypes></Header><Snippet>";

        private readonly string snippetXml3 = "<Code Language=\"csharp\"><![CDATA[";

        private readonly string snippetXml4 = "]]></Code></Snippet></CodeSnippet></CodeSnippets>";

        #endregion

        private ICommand _removeCommand;
        private ICommand _addCommand;
        private ICommand _cancelCommand;
        private ICommand _pathCommand;
        private ICommand _saveCommand;
        private ICommand _searchCommand;
        private ICommand _selectedItemChangedCommand;
        private ICommand _snippetPathCommand;
        private ICommand _snippetSaveCommand;
        private ICommand _fileSearchCommand;
        private ICommand _openFileCommand;
        private ICommand _saveFileCommand;

        private ISnippetService _service;
        private IDialogCoordinator _dialogCoordinator;
        public MainWindowViewModel(ISnippetService service, IDialogCoordinator dialogCoordinator)
        {
            _service = service;
            _dialogCoordinator = dialogCoordinator;
        }

        #region 生成本地代码

        private string _snippet = "";

        private string _snippetTitle;
        public string SnippetTitle
        {
            get => _snippetTitle;
            set
            {
                Set(ref _snippetTitle, value);
            }
        }

        private string _snippetAuthor;
        public string SnippetAuthor
        {
            get => _snippetAuthor;
            set => Set(ref _snippetAuthor, value);
        }

        private string _snippetShortcut;
        public string SnippetShortcut
        {
            get => _snippetShortcut;
            set
            {
                Set(ref _snippetShortcut, value);
            }
        }

        private string _snippetDescription;
        public string SnippetDescription
        {
            get => _snippetDescription;
            set => Set(ref _snippetDescription, value);
        }

        private TextDocument _snippetCode = new();
        public TextDocument SnippetCode
        {
            get => _snippetCode;
            set => Set(ref _snippetCode, value);
        }

        private List<Snippets> _snippetsList;
        public List<Snippets> SnippetsList
        {
            get => _snippetsList;
            set => Set(ref _snippetsList, value);
        }

        private string _snippetsPath;
        public string SnippetsPath
        {
            get => _snippetsPath;
            set => Set(ref _snippetsPath, value);
        }

        private bool _isBusy;

        private bool IsBusy
        {
            get => _isBusy;
            set => Set(ref _isBusy, value);
        }

        public ICommand SearchCommand
        {
            get { return _searchCommand ??= new DelegateCommand(CanExecute, SearchExecute); }
        }

        public ICommand SelectedItemChangedCommand
        {
            get { return _selectedItemChangedCommand ??= new DelegateCommand(CanExecute, SelectItemExecute); }
        }

        public ICommand SnippetPathCommand
        {
            get { return _snippetPathCommand ??= new DelegateCommand(CanExecute, SnippetsPathExecute); }
        }

        public ICommand SnippetSaveCommand
        {
            get { return _snippetSaveCommand ??= new AsyncCommand(SnippetSaveExecute, CanExecuteTask); }
        }

        // 执行保存
        private async Task SnippetSaveExecute()
        {
            IsBusy = true;

            // 路径不存在则返回
            if (string.IsNullOrWhiteSpace(SnippetsPath))
            {
                await _dialogCoordinator.ShowMessageAsync(this, "警告", "保存路径为空！");
                IsBusy = false;
                return;
            }

            // 内容不存在则返回
            if (string.IsNullOrWhiteSpace(_snippet))
            {
                await _dialogCoordinator.ShowMessageAsync(this, "警告", "保存内容为空！");
                IsBusy = false;
                return;
            }

            // 判断文件是否存在
            if (File.Exists(SnippetsPath + $"{SnippetTitle}.snippet"))
            {
                await _dialogCoordinator.ShowMessageAsync(this, "警告", $"名称为{SnippetTitle}的文件已存在，请重新命名！");
                IsBusy = false;
                return;
            }

            // 加载数据
            var document = new XmlDocument();
            document.LoadXml(_snippet);

            // 替换数据
            var manager = new XmlNamespaceManager(document.NameTable);
            manager.AddNamespace("ns", "http://schemas.microsoft.com/VisualStudio/2005/CodeSnippet");

            var titleNode = document.SelectSingleNode("//ns:Title", manager);
            var shortCutNode = document.SelectSingleNode("//ns:Shortcut", manager);

            if (titleNode != null) titleNode.InnerText = SnippetTitle;
            if (shortCutNode != null) shortCutNode.InnerText = SnippetShortcut;

            // 保存数据
            var fileStream = new FileStream(SnippetsPath + $"{SnippetTitle}.snippet", FileMode.OpenOrCreate, FileAccess.ReadWrite);
            var writer = new StreamWriter(fileStream);
            await writer.WriteLineAsync(document.InnerXml);
            writer.Close();

            // 提示保存成功
            await _dialogCoordinator.ShowMessageAsync(this, "信息", "保存成功");
            IsBusy = false;
        }

        // 执行路径命令
        private void SnippetsPathExecute(object parameter)
        {
            var dialog = new VistaFolderBrowserDialog();
            if (dialog.ShowDialog() == true)
            {
                SnippetsPath = $"{dialog.SelectedPath}\\";
            }
        }

        // 执行查询命令
        private void SearchExecute(object parameter)
        {
            // 若输入为空，则返回
            var searchString = parameter as string;
            if (string.IsNullOrWhiteSpace(searchString))
            {
                SnippetsList = null;
                return;
            }

            // 若输入为"*"查询所有数据
            SnippetsList = searchString switch
            {
                "*" => _service.GetSnippets(),
                _ => _service.GetSnippets(searchString)
            };
        }

        // 执行选择命令
        private void SelectItemExecute(object parameter)
        {
            var snippets = (Snippets)parameter;

            SnippetTitle = snippets == null ? "" : snippets.Title;
            SnippetAuthor = snippets == null ? "" : snippets.Author;
            SnippetShortcut = snippets == null ? "" : snippets.Shortcut;
            SnippetDescription = snippets == null ? "" : snippets.Description;
            SnippetCode.Text = snippets == null ? "" : snippets.Code;
            _snippet = snippets == null ? "" : snippets.Snippet;
        }

        private bool CanExecuteTask()
        {
            return !IsBusy;
        }

        #endregion

        #region 写入数据库

        private ObservableCollection<Declarations> _declarationsList = new();
        public ObservableCollection<Declarations> DeclarationsList
        {
            get => _declarationsList;
            set => Set(ref _declarationsList, value);
        }

        private string _title;
        public string Title
        {
            get => _title;
            set => Set(ref _title, value);
        }

        private string _author;
        public string Author
        {
            get => _author;
            set => Set(ref _author, value);
        }

        private string _shortcut;
        public string Shortcut
        {
            get => _shortcut;
            set => Set(ref _shortcut, value);
        }

        private string _description;
        public string Description
        {
            get => _description;
            set => Set(ref _description, value);
        }

        private TextDocument _code = new();
        public TextDocument Code
        {
            get => _code;
            set => Set(ref _code, value);
        }

        private string _oldId;
        public string OldId
        {
            get => _oldId;
            set => Set(ref _oldId, value);
        }

        private string _id;
        public string Id
        {
            get => _id;
            set => Set(ref _id, value);
        }

        private string _default;
        public string Default
        {
            get => _default;
            set => Set(ref _default, value);
        }

        private string _toolTip;
        public string ToolTip
        {
            get => _toolTip;
            set => Set(ref _toolTip, value);
        }

        private string _errorMessage;
        public string ErrorMessage
        {
            get => _errorMessage;
            set => Set(ref _errorMessage, value);
        }

        private string _filePath;
        public string FilePath
        {
            get => _filePath;
            set => Set(ref _filePath, value);
        }

        // 取消命令
        public ICommand CancelCommand
        {
            get { return _cancelCommand ??= new DelegateCommand(CanExecute, CancelExecute); }
        }

        // 添加命令
        public ICommand AddCommand
        {
            get { return _addCommand ??= new DelegateCommand(CanExecute, AddExecute); }
        }

        // 路径命令
        public ICommand PathCommand
        {
            get { return _pathCommand ??= new DelegateCommand(CanExecute, PathExecute); }
        }

        // 保存命令
        public ICommand SaveCommand
        {
            get { return _saveCommand ??= new AsyncCommand(SaveExecute, CanExecuteTask); }
        }

        // 删除命令
        public ICommand RemoveCommand
        {
            get { return _removeCommand ??= new DelegateCommand(CanExecute, RemoveExecute); }
        }

        // 执行取消命令
        private void CancelExecute(object parameter)
        {
            // 清空数据
            OldId = Id = Default = ToolTip = string.Empty;
        }

        // 执行添加命令
        private void AddExecute(object parameter)
        {
            // 判断数据是否全部填写
            if (string.IsNullOrWhiteSpace(OldId) || string.IsNullOrWhiteSpace(Id) || string.IsNullOrWhiteSpace(Default) || string.IsNullOrWhiteSpace(ToolTip))
            {
                ErrorMessage = "所有信息必須都填写！";
                return;
            }

            // 添加数据
            DeclarationsList.Add(new Declarations()
            {
                Guid = new Guid(),
                OldId = OldId,
                Id = Id,
                Default = Default,
                ToolTip = ToolTip
            });

            // 清空数据
            ErrorMessage = OldId = Id = Default = ToolTip = string.Empty;
        }

        // 获取路径
        private void PathExecute(object parameter)
        {
            var dialog = new VistaFolderBrowserDialog();
            if (dialog.ShowDialog() == true)
            {
                FilePath = $"{dialog.SelectedPath}\\";
            }
        }

        // 执行保存命令
        private async Task SaveExecute()
        {
            IsBusy = true;

            // 若存在信息为空则返回
            var codeText = Code.Text;
            if (string.IsNullOrWhiteSpace(Title) || string.IsNullOrWhiteSpace(Author) || string.IsNullOrWhiteSpace(Shortcut) || string.IsNullOrWhiteSpace(Description) || string.IsNullOrWhiteSpace(codeText))
            {
                IsBusy = false;
                return;
            }

            // 若存在参数，则替换参数
            string literalString = "";
            if (DeclarationsList.Count > 0)
            {
                foreach (var declarations in DeclarationsList)
                {
                    codeText = codeText.Replace(declarations.OldId, $"${declarations.Id}$");
                    literalString += $"<Literal><ID>{declarations.Id}</ID><Default>{declarations.Default}</Default><ToolTip>{declarations.ToolTip}</ToolTip></Literal>";
                }

                literalString = $"<Declarations>{literalString}</Declarations>";
            }

            // 拼接需要保存的数据
            var snippetString = $"{snippetXml1}<Title>{Title}</Title><Shortcut>{Shortcut}</Shortcut><Description>{Description}</Description><Author>{Author}</Author>{snippetXml2}{literalString}{snippetXml3}{codeText}{snippetXml4}";

            // 格式化xml，方便获取时显示
            var xmlCode = FormatXml(snippetString);

            // 准备保存的数据
            var guid = new Guid();
            var snippets = new Snippets()
            {
                Guid = guid,
                Title = Title,
                Author = Author,
                Shortcut = Shortcut,
                Description = Description,
                Code = Code.Text,
                Snippet = xmlCode,
                Declarations = DeclarationsList
            };

            // 保存本地
            if (File.Exists(FilePath + $"{Title}.snippet"))
            {
                await _dialogCoordinator.ShowMessageAsync(this, "警告", $"名称为{Title}的文件已存在，请重新命名！");
                IsBusy = false;
                return;
            }

            var fileStream = new FileStream(FilePath + $"{Title}.snippet", FileMode.OpenOrCreate, FileAccess.ReadWrite);
            var writer = new StreamWriter(fileStream);
            await writer.WriteLineAsync(xmlCode);
            writer.Close();

            // 保存数据
            _service.Save(snippets);

            // 提示保存成功
            await _dialogCoordinator.ShowMessageAsync(this, "消息", "保存成功！");

            // 保存完毕清空数据
            Title = Author = Shortcut = Description = Code.Text = string.Empty;
            DeclarationsList = null;

            IsBusy = false;
        }

        // 执行删除命令
        private void RemoveExecute(object parameter)
        {
            // 获取需要删除的行，从集合中删除
            var index = DeclarationsList.IndexOf(parameter as Declarations);
            if (index > -1 && index < DeclarationsList.Count)
            {
                DeclarationsList.RemoveAt(index);
            }
        }

        // 是否允许执行命令
        private bool CanExecute(object parameter)
        {
            return true;
        }

        public string Error => string.Empty;

        // 提示不能为空
        public string this[string columnName]
        {
            get
            {
                switch (columnName)
                {
                    case nameof(Title) when string.IsNullOrWhiteSpace(Title):
                    case nameof(Author) when string.IsNullOrWhiteSpace(Author):
                    case nameof(Shortcut) when string.IsNullOrWhiteSpace(Shortcut):
                    case nameof(Description) when string.IsNullOrWhiteSpace(Description):
                        return "不能为空";
                    default:
                        return null;
                }
            }
        }

        #endregion

        #region 修改本地代码

        private string _localCode;

        private string _fileName;
        public string FileName
        {
            get => _fileName;
            set => Set(ref _fileName, value);
        }

        private XmlDataProvider _xmlData = new();
        public XmlDataProvider XmlData
        {
            get => _xmlData;
            set => Set(ref _xmlData, value);
        }

        private string _localTitle;
        public string LocalTitle
        {
            get => _localTitle;
            set => Set(ref _localTitle, value);
        }

        private string _localAuthor;
        public string LocalAuthor
        {
            get => _localAuthor;
            set => Set(ref _localAuthor, value);
        }

        private string _localShortcut;
        public string LocalShortcut
        {
            get => _localShortcut;
            set => Set(ref _localShortcut, value);
        }

        private string _localDes;
        public string LocalDes
        {
            get => _localDes;
            set => Set(ref _localDes, value);
        }

        private ObservableCollection<Literal> _literals = new();
        public ObservableCollection<Literal> Literals
        {
            get => _literals;
            set => Set(ref _literals, value);
        }

        public ICommand FileSearchCommand
        {
            get { return _fileSearchCommand ??= new DelegateCommand(CanExecute, FileSearchExecute); }
        }

        public ICommand OpenFileCommand
        {
            get { return _openFileCommand ??= new AsyncCommand(OpenFileExecute, CanExecuteTask); }
        }

        public ICommand SaveFileCommand
        {
            get { return _saveFileCommand ??= new AsyncCommand(SaveFileExecute, CanExecuteTask); }
        }

        private void FileSearchExecute(object parameter)
        {
            var fileDialog = new VistaOpenFileDialog { Filter = "打开代码块 (*.snippet)|*.snippet" };
            if (fileDialog.ShowDialog() == true)
            {
                FileName = fileDialog.FileName;
            }
        }

        private async Task OpenFileExecute()
        {
            IsBusy = true;

            // 判断文件是否存在
            if (!File.Exists(FileName))
            {
                await _dialogCoordinator.ShowMessageAsync(this, "警告", "文件不存在，请检查！");
                IsBusy = false;
                return;
            }

            // 读取数据
            var stream = new FileStream(FileName, FileMode.Open, FileAccess.Read);
            var bytes = new byte[stream.Length];
            var num = await stream.ReadAsync(bytes, 0, bytes.Length);
            var str = Encoding.Default.GetString(bytes, 0, num);

            // 显示到界面
            var document = new XmlDocument();
            document.LoadXml(str);

            // 解析xml
            var manager = new XmlNamespaceManager(document.NameTable);
            manager.AddNamespace("ns", "http://schemas.microsoft.com/VisualStudio/2005/CodeSnippet");

            LocalTitle = document.SelectSingleNode("//ns:Title", manager)?.InnerText;
            LocalAuthor = document.SelectSingleNode("//ns:Author", manager)?.InnerText;
            LocalShortcut = document.SelectSingleNode("//ns:Shortcut", manager)?.InnerText;
            LocalDes = document.SelectSingleNode("//ns:Description", manager)?.InnerText;
            _localCode = document.SelectSingleNode("//ns:Code ", manager)?.InnerText;

            var xmlNode = document.SelectSingleNode("//ns:Declarations", manager);
            if (xmlNode?.ChildNodes.Count > 0)
            {
                Literals = new ObservableCollection<Literal>();
                foreach (XmlNode node in xmlNode.ChildNodes)
                {
                    Literals.Add(new Literal
                    {
                        ID = node.SelectSingleNode("//ns:ID", manager)?.InnerText,
                        Default = node.SelectSingleNode("//ns:Default", manager)?.InnerText,
                        ToolTip = node.SelectSingleNode("//ns:ToolTip", manager)?.InnerText
                    });
                }
            }

            XmlData.Document = document;
            XmlData.XPath = "*";

            IsBusy = false;
        }

        private async Task SaveFileExecute()
        {
            IsBusy = true;

            // 判断所输入信息是否为空，为空则返回
            if (string.IsNullOrWhiteSpace(LocalTitle) || string.IsNullOrWhiteSpace(LocalShortcut))
            {
                IsBusy = false;
                return;
            }

            // 判断文件是否存在
            var path = FileName.Substring(0, FileName.LastIndexOf('\\') + 1);
            if (File.Exists(path + $"{LocalTitle}.Snippet"))
            {
                await _dialogCoordinator.ShowMessageAsync(this, "警告", $"名称为{Title}的文件已存在，请重新命名！");
                IsBusy = false;
                return;
            }

            // 若存在参数，拼接参数
            string literalString = "";
            if (Literals != null && Literals.Count > 0)
            {
                foreach (var literal in Literals)
                {
                    literalString += $"<Literal><ID>{literal.ID}</ID><Default>{literal.Default}</Default><ToolTip>{literal.ToolTip}</ToolTip></Literal>";
                }

                literalString = $"<Declarations>{literalString}</Declarations>";
            }

            // 拼接需要保存的数据
            var snippetString = $"{snippetXml1}<Title>{LocalTitle}</Title><Shortcut>{LocalShortcut}</Shortcut><Description>{LocalDes}</Description><Author>{LocalAuthor}</Author>{snippetXml2}{literalString}{snippetXml3}{_localCode}{snippetXml4}";

            // 格式化xml，方便获取时显示
            var xmlCode = FormatXml(snippetString);

            // 删除现在的文件
            File.Delete(FileName);

            // 重新写入
            var fileStream = new FileStream(path + $"{LocalTitle}.snippet", FileMode.OpenOrCreate, FileAccess.ReadWrite);
            var writer = new StreamWriter(fileStream);
            await writer.WriteLineAsync(xmlCode);
            writer.Close();

            // 提示保存成功
            await _dialogCoordinator.ShowMessageAsync(this, "信息", "保存成功");

            // 清空数据
            FileName = LocalTitle = LocalAuthor = LocalDes = LocalShortcut = string.Empty;
            XmlData.Document = null;
            Literals = null;

            IsBusy = false;
        }

        #endregion

        private static string FormatXml(string xmLString)
        {
            //校验是否是XML报文
            if (!xmLString.Contains("<?xml version")) return xmLString;
            XmlDocument xmlDocument = GetXmlDocument(xmLString);
            return ConvertXmlDocumentToString(xmlDocument);
        }

        private static XmlDocument GetXmlDocument(string xmlString)
        {
            XmlDocument document = new XmlDocument();
            document.LoadXml(xmlString);
            return document;
        }

        // 格式化Xml
        private static string ConvertXmlDocumentToString(XmlDocument xmlDocument)
        {
            var memoryStream = new MemoryStream();
            var writer = new XmlTextWriter(memoryStream, null)
            {
                Formatting = Formatting.Indented//缩进
            };
            xmlDocument.Save(writer);
            var streamReader = new StreamReader(memoryStream);
            memoryStream.Position = 0;
            var xmlString = streamReader.ReadToEnd();
            streamReader.Close();
            memoryStream.Close();
            return xmlString;
        }
    }
}
